#!/bin/sh

echo -e "\e[34m=== tools-build.sh"

# Set up the environment to a known state.

SRCDIR=`pwd`/sources
set +h
umask 022
export LC_ALL=POSIX
export CLFAGS="-Os"
export PATH=/sbin:/usr/sbin:/bin:/usr/bin
export HOME=/root
export TERM=linux
unset CC CXX CPP LD_LIBRARY_PATH LD_PRELOAD
# There must be a better way to do this one.
ARCH=i386

########## Phase 1-1: Use parent system to build a uclibc cross-compiler

# First, we use the parent system (assumed to be some reasonably sane Linux
# variant) to build a working set of cross-compile build tools.  These tools
# run on the parent system (which could be glibc), but produce binaries linked
# against the uclibc in /tools/lib.  These build tools include a c compiler
# (gcc), a linker and assembler (binutils), and a C library (uclibc) to link
# the resulting binaries against.  Building and using the C library also
# requires a set of kernel ABI headers (matching the linux kernel version to
# be installed on the destination system).

# A couple of symlinks to make binutils and gcc builds happy.

mkdir -p /tools/{bin,etc,sources,usr} &&
#ln -s /lib /tools/lib &&
#ln -s /usr/lib /tools/usr/lib &&
#ln -s /usr/include /tools/include

# Unpack linux kernel headers somewhere uclibc can find them.

cd /tools/sources &&
tar xvjf "$SRCDIR/buildtools/linux-libc-headers-2.6.12.0.tar.bz2" &&
cd linux-libc-headers-2.6.12.0/include &&
mv asm-$ARCH ../../asm &&
mv linux sound ../.. &&
cd ../.. &&
rm -rf linux-libc-headers-2.6.12.0

if [ $? -ne 0 ]; then exit 1; fi

echo === Build uclibc

cd /tools/sources &&
tar xvjf "$SRCDIR/system/uClibc-0.9.28.tar.bz2" &&
cd uClibc-0.9.28 &&
patch -p1 -i "$SRCDIR/system/uClibc-0.9.28-libc_stack_end-1.patch" &&
bzcat "$SRCDIR/system/uClibc-config.bz2" > .config &&
# fix_includes gets deeply confused by pre-prepared kernel headers, so
# do its job manually.
echo > extra/scripts/fix_includes.sh &&
mv ../asm ../linux ../sound include &&
make &&
# Install's a bit funky, but this seems to work.
make RUNTIME_PREFIX=/tools/ DEVEL_PREFIX=/tools/ install_runtime install_dev &&
# make clean breaks if asm, linux, and scsi aren't symlinks...
#mv include/asm include/linux include/scsi . &&
#make clean &&
#mv asm linux scsi include

if [ $? -ne 0 ]; then exit 1; fi

echo "=== Build binutils that links against /tools"

cd /tools/sources &&
tar xvjf "$SRCDIR/buildtools/binutils-2.16.1.tar.bz2" &&
sed -i -e 's@^NATIVE_LIB_DIRS.*@NATIVE_LIB_DIRS=/tools/lib@' binutils-2.16.1/ld/configure.tgt &&
mkdir binutils-build &&
cd binutils-build &&
../binutils-2.16.1/configure --prefix=/tools --disable-nls &&
make configure-host &&
make &&
make install &&
cd .. &&
rm -rf binutils-2.16.1 binutils-build

if [ $? -ne 0 ]; then exit 1; fi

echo "=== Build gcc that links against /tools"
# Note: we need to modify gcc source a LOT, both to point various paths at
# the /tools directory, and to point them at uclibc instead of glibc.
# In theory we should be able to tell ./configure "--prefix=/tools" and
# "--arch=$(ARCH)-pc-linux-uclibc", and have everything else just work.
# Unfortunately, this doesn't actually work...

cd /tools/sources &&
tar xvjf "$SRCDIR/buildtools/gcc-core-4.0.2.tar.bz2" &&
cd gcc-4.0.2 &&
# Do NOT run fixincludes (Replaces no_fixincludes patch from LFS)
sed -i -e 's@\./fixinc\.sh@-c true@' -e 's@.*include/README@	echo@' \
	gcc/Makefile.in &&
## Fix pointlessly hardwired paths.
#sed -i 's@\(standard_startfile_prefix_[12].*\)"/@&tools/@g' gcc/gcc.c &&
# Remove pointlessly hardwired paths.
sed -i -e 's@(\*standard_startfile_prefix_[12])@(0)@' gcc/gcc.c &&
# Edit other pointlessly hardwired paths
sed -r -i -e 's@.*"/usr/lib".*@@' -e 's@.*"/usr/local/lib".*@@' -e 's@"/lib@"/tools/lib@g' -e 's@"/usr@"/tools/usr@g' gcc/collect2.c &&
# Adjust config entries from which spec file will be generated
for i in `find gcc/config -name "*.h"`
do
  sed -i -r -e 's /lib(64)?/ld(64)?(-linux)?(-x86-)?(-ia)?(64)?.so.[12] /tools/lib/ld-uClibc.so.0 ' -e 's P,/usr/lib P,/tools/lib ' -e 's@-rpath /lib@-rpath /tools/lib@' "$i"
  if [ $? -ne 0 ]; then exit 1; fi
done &&
# Remove /usr/include from the end of the include search path.
# (Also part of specs-2 patch from LFS)
cat >> gcc/config/linux.h << EOF &&
#undef STANDARD_INCLUDE_DIR
#define STANDARD_INCLUDE_DIR 0
EOF
# Chop out a stack unwinding symbol uClibc doesn't provide.
#sed -i -e 's/(dl_iterate_phdr.*< 0)/(0)/' gcc/unwind-dw2-fde-glibc.c &&
# We don't need this, and it would break for the second build, so just yank it.
rm -rf libmudflap &&
# We've adjusted the source, now build it.
cd .. &&
mkdir gcc-build &&
cd gcc-build &&
../gcc-4.0.2/configure --prefix=/tools --with-local-prefix=/tools --disable-nls --enable-shared --enable-languages=c &&
# GCC's normal build process (make bootstrap) actually builds gcc three times:
# first with the parent system's compiler, then rebuilds it using the compiler
# it built in the first pass, and finally builds a third instance using the
# second one to make sure they produce identical binaries.  (This level of
# paranoia comes from cross-compiling gcc from Solaris or HP-UX using the
# system's wonky native compilers.)
#
# The "quickstrap" target just builds the first pass and installs it.
# We're building from a Linux system, and we're going to recompile it with
# itself anyway in phase 1-2.
make quickstrap &&
make install &&
ln -sf gcc /tools/bin/cc &&
cd .. &&
rm -rf gcc-build gcc-4.0.2

if [ $? -ne 0 ]; then exit 1; fi

# Remove temporary symlinks needed by above builds.
#rm -rf /tools/lib /tools/usr/lib /tools/usr /tools/include &&

# The phase 1 toolset is complete and installed.  All we have to do to use
# it is put /tools/bin at the start of the path, so the binaries it contains
# are found before the ones in the host system.
export "PATH=/tools/bin:$PATH"
# Quick test to make sure the toolset is creating binaries linked against
# /tools/lib and uclibc.
echo 'main(){}' > dummy.c &&
gcc dummy.c &&
xtest=`readelf -l a.out | grep ': /tools/lib'` &&
if [ -z "$xtest" ]
then
  echo "error: Toolchain isn't linking against /tools directory"
  exit 1
fi
# Now make sure there's nothing _but_ uclibc getting linked in.
xtest=`ldd a.out | grep -v "libc.so.0" | grep -v "ld-uClibc.so.0"`
if [ -z $xtest ]
then
  echo "Toolchain diagnostics check out."
else
  echo "error: Toolchain is not creating uclibc-only dynamically linked executables."
  exit 1
fi
unset xtest
rm -f dummy.c a.out

# Phase 2: Add utilities.
#
# Everything we build with the new toolchain will depend on /tools/lib, so
# we can drop a nice self-contained /tools directory into an otherwise empty
# directory, chroot into the empty direcytory to build stuff, and then cleanly
# remove /tools when we're done.
#
# A compiler alone isn't enough to build stuff with.  In theory adding busybox
# and make should suffice.  (In practice, busybox isn't quite there yet, so
# we need bash too.)
#
# Since we still have the uClibc source lying around, let's start with that.
# We don't have to reinstall uClibc (even though this version was built by
# the parent system's compiler, it'll get rebuilt anyway for the final
# install.)  But we do need versions of ldd and such that link against /tools.

echo === Build uclibc utilities. &&
cd /tools/sources/uClibc-0.9.28 &&
# We ran the library build earlier, so the includes directory should have
# the correct uClibc headers in it.
make RUNTIME_PREFIX=/tools/ DEVEL_PREFIX=/tools/ install_utils &&
cd .. &&
rm -rf uClibc-0.9.28

if [ $? -ne 0 ]; then exit 1; fi

echo === busybox-1.1.0-pre1

cd /tools/sources &&
tar xvjf "$SRCDIR/system/busybox-20051018.tar.bz2" &&
cd busybox &&
bzcat "$SRCDIR/system/busybox-config.bz2" > .config &&
make &&
cp busybox /tools/bin/ &&
rm /tools/bin/strings && # Use busybox version, not Binutils version.
(for i in `sed 's ^.*/  ' busybox.links`; do ln -s busybox "/tools/bin/$i"; done) &&
# Can't use patch sh ar...  Fix all that later.
cd .. &&
rm -rf busybox

if [ $? -ne 0 ]; then exit 1; fi

echo === make-3.80

cd /tools/sources &&
tar xvjf "$SRCDIR/buildtools/make-3.80.tar.bz2" &&
cd make-3.80 &&
./configure --prefix=/tools &&
make &&
make install &&
cd .. &&
rm -rf make-3.80

if [ $? -ne 0 ]; then exit 1; fi

# Someday, the three busybox shells should be unified and extended into a
# workable bash replacement.  Until then, we supplement busybox with bash.

echo === bash

cd /tools/sources &&
tar xvzf "$SRCDIR/temp/bash-2.05b.tar.gz" &&
cd bash-2.05b &&
cat "$SRCDIR"/temp/bash-*patch | patch -p1 &&
# Fix bash's default paths (mostly because "/usr/gnu/bin" is just stupid).
sed -i 's@".*/usr/bin.*"@"/tools/bin"@' config-top.h &&
./configure --prefix=/tools &&
make &&
make install &&
cd .. &&
rm -rf bash-2.05b &&
ln -s bash "/tools/bin/sh"

if [ $? -ne 0 ]; then exit 1; fi

# Phase 3: Rebuild binutils and gcc
#
# Problem 1: The gcc and binutils binaries we built earlier are linked against
# the parent system.  Although they generate stuff that can work when /tools
# stands alone in a chroot environment, they themselves can't run from there.
#
# Problem 2: When we use /tools to build a new environment, we want the new
# binaries we build to work without /tools.
#
# Building a new compiler with the current compiler will fix problem 1,
# and we can configure the new compiler to fix problem 2.

echo === rebuild binutils dynamically linked against uclibc.

cd /tools/sources &&
tar xvjf "$SRCDIR/buildtools/binutils-2.16.1.tar.bz2" &&
mkdir binutils-build &&
cd binutils-build &&
../binutils-2.16.1/configure --prefix=/tools --enable-shared && #--with-lib-path=/tools/lib &&
make NATIVE_LIB_DIRS=/lib &&
make install &&
cd .. &&
rm -rf binutils-build binutils-2.16.1

if [ $? -ne 0 ]; then exit 1; fi

echo === Rebuild gcc with /tools, dynamically linked against uclibc.

cd /tools/sources &&
tar xvjf "$SRCDIR/buildtools/gcc-core-4.0.2.tar.bz2" &&
cd gcc-4.0.2 &&
# Do NOT run fixincludes (Replaces no_fixincludes patch from LFS)
sed -i -e 's@\./fixinc\.sh@-c true@' -e 's@.*include/README@	echo@' \
	gcc/Makefile.in &&
# Remove pointlessly hardwired paths.
sed -i -e 's@(\*standard_startfile_prefix_[12])@(0)@' gcc/gcc.c &&
# Libmudflap will try to build using the new compiler before it's even
# installed, meaning # it'll complain there's no /lib/uClibc* yet.  Well duh.
rm -rf libmudflap &&
# Don't install redundant libiberty
sed -i "s/install:.*install-subdir/install: install-subdir/" libiberty/Makefile.in &&
for i in `find gcc/config -name "*.h"`
do
  sed -i -r -e 's /lib(64)?/ld(64)?(-linux)?(-x86-)?(-ia)?(64)?.so.[12] /lib/ld-uClibc.so.0 ' "$i"
  if [ $? -ne 0 ]; then exit 1; fi
done &&
mkdir ../gcc-build &&
cd ../gcc-build &&
../gcc-4.0.2/configure --prefix=/tools --with-local-prefix=/tools --disable-nls --enable-shared --enable-languages=c &&
#../gcc-4.0.2/configure --prefix=/tools --with-local-prefix=/tools --enable-shared --enable-threads=posix --enable-__cxa_atexit --enable-clocale=gnu --enable-languages=c &&
# Once again, build just once rather than three times.  It's already being
# built with itself, and we'll build it again later in the final system.
make quickstrap &&
make install &&
cd .. &&
rm -rf gcc-build gcc-4.0.2

if [ $? -ne 0 ]; then exit 1; fi

# Now that we're done building stuff linked against /tools, adjust linker to
# build stuff linked against /lib instead.

for i in /tools/*linux*/lib/ldscripts/*
do
sed -i -e 's@SEARCH_DIR("/tools/lib")@SEARCH_DIR("/lib")@' "$i"
done
sed -i -e 's@ /tools/lib/ld-uClibc.so.0@ /lib/ld-uClibc.so.0@g' \
        /tools/lib/gcc-lib/*/*/specs &&

echo === strip debugging symbols.

rm -rf /tools/{doc,info,man} &&
find /tools/lib -type f | xargs strip --strip-debug &&
for i in /tools/{,s}bin; do (find "$i" -type f | xargs strip --strip-unneeded ); done

# Let parent script know we completed successfully.
touch /tools/finished


